#|
Copyright (c) 2006 Risto Laakso
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|#
(in-package :event-notification)

#+linux
(progn

  (defconstant +epoll-size+ 1000)

  (defun make-event-queue ()
    (%epoll-create +epoll-size+))

  (defun read-event-p (event)
    (not (eql (logand event +epoll-in+) 0)))

  (defun write-event-p (event)
    (not (eql (logand event +epoll-out+) 0)))



  (defun add-fd (event-queue fd mode &key (trigger :edge))
    (with-foreign-object (ev 'epoll-event)
      (memzero ev +epoll-event-size+)
      (let ((actual-mode (logior (if (eql :read mode) +epoll-in+ 0) 
				 (if (eql :write mode) +epoll-out+ 0)
				 (if (eql :read-write mode) (logior +epoll-in+ +epoll-out+) 0)
				 (if (eql trigger :edge) +epoll-et+ 0))))
	#+nio-debug	(format t "Add-fd called with :fd ~A :event-queue ~A :mode ~A :trigger ~A :actual-mode ~A~%" fd event-queue mode trigger actual-mode)
	
	(setf (foreign-slot-value ev 'epoll-event 'fd)  fd
	      (foreign-slot-value ev 'epoll-event 'events)
	      actual-mode))
      
      (if (eql (%epoll-ctl event-queue +epoll-ctl-add+ fd ev) -1)
	  (progn
	    (format t "add-fd (epoll_ctl) error occurred: ~A~%" (get-errno))
	    ;;	    (error 'poll-error)
	    ))))
  
  (defun remove-fd (event-queue fd mode)
    (with-foreign-object (ev 'epoll-event)
      (memzero ev +epoll-event-size+)

      (setf (foreign-slot-value ev 'epoll-event 'fd) fd)

      (%epoll-ctl event-queue +epoll-ctl-del+ fd ev)))


  (define-condition poll-error (error) ())

  (defun poll-events (event-queue)
#+nio-debug    (format t "poll-events called with :event-queue ~A~%" event-queue)
    (with-foreign-object (events 'epoll-event +epoll-size+)
      (memzero events (* +epoll-event-size+ +epoll-size+))
      (loop for res = (%epoll-wait event-queue events +epoll-size+ -1)
	   
	   do
	   (progn
#+nio-debug	     (format t "poll-events - dealing with ~A~%" res)
	     (case res
	       (-1 (error 'poll-error))
	       (0 nil)
	       (t 
		(let ((idents nil))
		  (loop for i from 0 below res do
		       (push (cons (foreign-slot-value 
				    (mem-aref events 'epoll-event i) 
				    'epoll-event 'fd)
				   (foreign-slot-value 
				    (mem-aref events 'epoll-event i) 
				    'epoll-event 'events)) 
			     idents))
		  (return idents))))))))
)
